!
!     Source Programs for calculating properties of cubic EOS
!     like Soave-Redlich-Kwong (SRK) and  Peng-Robinson (PR)
!
!     Copyright:   Michael L. Michelsen
!                  IVC-SEP, Dept. Chemical Engineering
!                  Bygning 229, DTU, DK 2800 Lyngby, DENMARK
!
      MODULE CUBIC_EOS
      INTEGER N,NDER,NTEMP
      DOUBLE PRECISION C1,C2,C
      INTEGER, ALLOCATABLE::          NOTRY(:)
      DOUBLE PRECISION, ALLOCATABLE:: AC(:),Q(:),TSQR(:),TCR(:),PCR(:)
      DOUBLE PRECISION, ALLOCATABLE:: OMG(:),BC(:),AC0(:),AC1(:),AC2(:)
      DOUBLE PRECISION, ALLOCATABLE:: ONEMK(:,:)
      CHARACTER*4, ALLOCATABLE::      NWNAME(:)
      PRIVATE
!
!     Allow only external access to initialization routine, calculation
!     routine and critical property routine
!
      PUBLIC THERMO,INI_CUB,GETCRIT
!
!     THERMODYNAMICS BLOCK
!
      CONTAINS
      SUBROUTINE THERMO(ICON,NTYP,MTYP,T,P,ZZ,Z,FG,FT,FP,FX,AUX)
!
!     Control routine for property calculation
!
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
      DIMENSION Z(*) , FG(*) , FT(*) , FP(*) , FX(:,:) , AUX(*)
!
!     PARAMETERS OF THERMO
!
!     ICON:  (I)   SETS PROPERTIES TO BE CALCULATED
!                  1  =  FG
!                  2  =  FG,FT,FP, H_RES, S_RES
!                  3  =  FG, FX
!                  4  =  FG,FT,FP, FX, H_RES, S_RES
!                  5  =  ALL + CP
!     NTYP:  (I)   1  = LIQUID
!                 -1  = VAPOUR
!                  0  = MIN. G-PHASE
!     MTYP:  (I)   RETURNED PHASE TYPE:
!                  NTYP = 1/-1:  MTYP = 1 IF FOUND,-1 IF NOT
!                  NTYP =    0:  MTYP = 2 IF LIQUID, -2 IF VAPOUR
!     T:     (I)   TEMPERATURE (K)
!     P:     (I)   PRESSURE (MPa)
!     ZZ:    (O)   COMPRESSIBILITY FACTOR
!     Z:     (I)   FEED MOLES, NORMALIZATION NOT NEEDED
!     FG:    (O)   VECTOR OF LN FUGACITY COEFFIICENTS
!     FT:    (O)   T-DERIVATIVES OF FG (IF REQUESTED)
!     FP:    (O)   P-DERIVATIVES OF FG (IF REQUESTED)
!     FX:    (O)   MATRIX OF SCALED N-DERIVATIVES
!     AUX:   (O)   OTHER PROPERTIES LIKE H_RES, S_RES, CP_RES
!
!     ***************************************************************
!     *                                                             *
!     *  Note: Size of AUX must be at least 9                       *
!     *        Size of other vector arguments at least NCOMP        *
!     *        Size of array arguments at least NCOMP x NCOMP       *
!     *                                                             *
!     *                                                             *
!     ***************************************************************
!
      NTEMP = 0
      NDER = 1
      IF ( ICON.GT.2 ) NDER = 2
      IF ( ICON.EQ.2 .OR. ICON.GT.3 .OR. ICON.EQ.0 ) NTEMP = 1
      IF ( ICON.GT.4 ) NTEMP = 2
      CALL CUBGEN(NTYP,MTYP,T,P,ZZ,Z,FG,FT,FP,FX,AUX)
      END SUBROUTINE
!
!
!
      SUBROUTINE CUBGEN(MT,IC,T,P,Z_TRUE,XX,FG,FT,FP,FX,AUX)
!
!     PARAMETERS:
!
!     MT:     (I):      PHASE TYPE DESIRED
!                       1 = LIQ, -1 = VAP, 0 = MIN G PHASE
!     IC:     (O):      INDICATOR FOR PHASE TYPE; IC RETURNS
!                       1:   IF CALLED WITH MT = 1/-1 AND PHASE IS FOUND
!                       -1:  IF CALLED WITH MT = 1/-1 AND PHASE IS NOT F
!                       2:   IF CALLED WITH MT = 0 AND MIN G PHASE IS LI
!                       -2:  IF CALLED WITH MT = 0 AND MIN G PHASE IS VA
!     T:      (I):      TEMPERATURE (K)
!     P:      (I):      PRESSURE (ATM)
!     Z_TRUE  (O):      CALCULATED COMPRESSIBILITY FACTOR
!     XX:     (I):      MIXTURE TOTAL COMPOSITION
!     FG:     (O):      LOG FUGACITY COEFFICIENT VECTOR
!     FT:     (O):      T-DERIVATIVE OF FG
!     FP:     (O):      P-DERIVATIVE OF FG
!     FX:     (O):      SCALED COMPOSITION DERIVATIVE OF FG
!     AUX:    (O):      VARIOUS RESIDUAL PROPERTIES
!
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
!-----DUMMY VARIABLES
      DIMENSION XX(*) , FG(*) , FT(*) , FP(*) , FX(:,:) , AUX(*)
!-----LOCAL VARIABLES
      DIMENSION X(N),PD(N),AD1(N),ADT(N)
      LOGICAL PSPEC
!
!     GET T-DEPENDENT TERMS
!
      CALL TEMSET(T)
!
!     GENERATE A NORMALIZED COMPOSITION IN COPY
!
      SUMR = 1.D0/SUM(XX(1:N))
      X = SUMR*XX(1:N)
!
!     ANEW CALLS MIXTURE PARAMETERS A AND B AND THEIR APPROPRIATE
!     DERIVATIVES
!
      CALL ANEW(X,A,AT,ATT,B,AD1,ADT)
!
!     P = 0 IS TAKEN TO BE A SPECIAL CASE WHERE THE VOLUME (OR RATHER
!     V/R) IS THE INPUT AND THE CORRESPONDING P IS CALCULATED
!     VALUES HERE ARE NOT PENELOUX CORRECTED
!
      PSPEC = .TRUE.
      IF  (P .EQ. 0.D0) THEN
         PSPEC = .FALSE.
         V = Z
         P = (1.D0/(V-B) - A/((V+C1*B)*(V+C2*B)))*T
         Z = POVERT*V
      ENDIF
      TREC = 1.D0/T
      POVERT = P*TREC
      PREC = 1.D0/P
      BREC = 1.D0/B
      APT = A*POVERT
      BPT = B*POVERT
      IF (PSPEC) THEN
!
!     SOLVE CUBIC EOS
!
          CALL CUBIC(MT,APT,BPT,Z,Z2,GDER)
!-------SECOND Z AND ENERGY DIFF. STORED IN AUX(1),AUX(2)
          AUX(1) = Z2
          Z_TRUE = Z
          AUX(2) = GDER
          ELSE
          AUX(1) = 0.D0
          AUX(2) = 0.D0
      ENDIF
!
!     WE TAKE - SOMEWHAT ARBITRARY - THE FLUID TO BE LIQUID IF V < 3*B
!
      V = Z*T*PREC
      ICX = 1
      IF (V .GT. 3.D0*B) ICX = -1
!
!     PHASE TYPE RETURN INDICATOR
!
      IF (MT.NE. 0) THEN
          IC = ICX*MT
          ELSE
          IC = 2*ICX
       ENDIF
!------V = VOLUME/R
!
!     COEFFICIENTS FOR EOS
!
      S1 = 1.D0/(V+C1*B)
      S2 = 1.D0/(V+C2*B)
!
!     N, A AND B DERIVATIVES OF P/T
!
      P1 = P*TREC + A*S1*S2
      PA = -S1*S2
      PN = P1
      FAC = C1*S1 + C2*S2
      P2 = A*PA
      PB = P1*P1 - FAC*P2
!
!       DERIVATIVE IS D(P/T) / D(V/R)
!
      DPDV = -P1*P1 - P2*(S1+S2)
      AUX(10) = DPDV
      AUX(8) = DPDV*T
!
!     COEFFICIENTS FOR FUGACITY COEFFICIENT CALCULATION
!     FA = DF/DA, FB = DF/DB, FAB = D2F/DADB ETC.
!
      FN  = LOG(V*P1)
      XL2 = LOG(S1/S2)/(C2-C1)
      FA = -XL2*BREC
      F2 = -A*FA
      FF = FN - F2
      GB = -V*PA
      F2B = (A*GB-F2)*BREC
      FB = P1 - F2B
      FNB = P1
      FAB = -F2B/A
      GBB = -GB*FAC
      F2BB = (A*GBB-2.D0*F2B)*BREC
      FBB = P1*P1 - F2BB
      XLZ = LOG(Z)
      FNP = FN - XLZ
!
!     ARE TEMPERATURE DERIVATIVES REQUIRED
!
      IF ( NTEMP.NE.0 ) THEN
         DFT = FA*AT
         HR = -T*DFT + Z - 1.D0
         HR = HR*T
!------RESID. ENTHALPY/R STORED IN AUX(3)
         AUX(3) = HR
         SR = -T*DFT - FF + XLZ
!------ RESID ENTROPY/R STORED IN AUX(4)
         AUX(4) = SR
         PTR = PA*AT
         DPDT = P*TREC + T*PTR
!
!     RESIDL CP REQUIRED
!
         IF ( NTEMP.GE.2 ) THEN
            FTT = FA*ATT
            CP = -T*(T*FTT+2.D0*DFT) - DPDT**2/DPDV - 1.D0
!------RESID HEAT CAPACITY/R IN AUX(5)
            AUX(5) = CP
            DVDT = -DPDT/DPDV*TREC
!---- AUX(6) IS PRESSURE DERIVATIVE OF RESID. ENTROPY
            AUX(6) = -DVDT + PREC
!-----AUX(7) IS PRESSURE DERIVATIVE OF ENTHALPY
            AUX(7) = -T*DVDT + V
            AUX(9) = DVDT
         ENDIF
      ENDIF
      IF ( NDER.NE.0 ) THEN
!
!     LOG FUGACITY COEFFICIENTS
!
         FG (1:N) = FNP + FA*AD1 + FB*BC
         IF ( NDER.GE.2 .OR. NTEMP.NE.0 ) THEN
            DPDVR = 1.D0/DPDV
            PD = DPDVR*(PN + PA*AD1 + PB*BC)
!
!     COMPOSITION DERIVATIVES REQUIRED
!
            IF ( NDER.GE.2 ) THEN
               FA2 = 2.D0*FA
               DO I = 1 , N
                  CC = 1.D0 + FNB*BC(I) + PN*PD(I)
                  CA = FAB*BC(I) + PA*PD(I)
                  CB = FNB + FAB*AD1(I) + FBB*BC(I) + PB*PD(I)
                  FAI = FA2*AC0(I)
                  FX(I:N,I) = CC + CA*AD1(I:)+CB*BC(I:)+                &
     &                          FAI*AC0(I:)*ONEMK(I:,I)
                  FX(I,I:N) = FX(I:N,I)
               ENDDO
            ENDIF
!
!     TEMPERATURE AND PRESSURE DERIVATIVES
!
            IF ( NTEMP.NE.0 ) THEN
               CB = FAB*AT
               DPDTT = DPDT*TREC
               FP(1:N) = -TREC*PD - PREC
               FT(1:N) = TREC + CB*BC + FA*ADT + DPDTT*PD
            ENDIF
         ENDIF
      ENDIF
      END SUBROUTINE
!
!     MIXTURE PARAMETERS A AND B
!
      SUBROUTINE ANEW(X,A0,AT,ATT,B0,AD1,ADT)
!
!     ANEW IS AN INTERNAL SUBROUTINE FOR TERMO
!     IT RETURNS THE B-PARAMETER, THE A-PARAMETER
!     AND THE REQUIRED COMPOSITION- AND TEMPERATURE
!     DERIVATIVES OF A
!
!
!     X:      (I):      NORMALIZED COMPOSITION
!     A0:     (O):      MIXTURE A (A/RT)-PARAMETER
!     AT:     (O):      T-DERIVATIVE OF A0
!     ATT:    (O):      T-DERIVATIVE OF AT
!     B0:     (O):      MIXTURE B-PARAMETER
!     AD1:    (O):      COMPSOTION DERIVATIVE OF A0
!     ADT:    (O):      T-DERIVATIVE OF AD1
!
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
      DIMENSION AD1(:),ADT(:),X(:)
      DIMENSION  AX0(N),AX1(N),AH0(N),AH1(N)
!------ HERE USED: SINGLE SUM IN B
      B0 = DOT_PRODUCT(X(1:N),BC)
      AX0 = X*AC0
      AX1 = X*AC1
!
!     SCALAR PRODUCT, SUM_J a_J X_J (1-K_IJ)
!
      AH0 = MATMUL (AX0,ONEMK)
      AD1  = 2.D0*AC0*AH0
      A0 = 0.5D0*DOT_PRODUCT (AD1,X)
      IF (NTEMP .GT. 0) THEN
          AH1 = MATMUL(AX1,ONEMK)
          ADT(1:N) = 2.D0*(AC0*AH1 + AC1*AH0)
          AT = 0.5D0*DOT_PRODUCT (ADT(1:N),X(1:N))
!
!     SECOND DERIVATIVE OF A NEEDED FOR RESIDUAL HEAT CAPACITY
!
          IF (NTEMP .GE. 2) THEN
              ATT  = 2.D0*DOT_PRODUCT(X,AC2*AH0+AC1*AH1)
          ENDIF
      ENDIF
      END SUBROUTINE
!
!
!
      SUBROUTINE CUBIC(MTYP,A,B,Z,ZV2,DELG)
!
!     CUBIC EQUATION SOLVER ROUTINE
!
!     MTYP:   (I):      DESIRED PHASE TYPE 1=LIQ,-1=VAP,0=MIN. G
!     A:      (I):      EOS AP/T
!     B:      (I):      EOS BP/T
!     Z:      (O):      RETURNED COMPRESSIBILITY FACTOR
!     ZV2:    (O):      SECOND ROOT, IF EXISTING
!     DELG    (O):      G/RT DIFFERENCE IN CASE OF MULTIPLE ROOTS
!
!------Z IS DESIRED ROOT; ZV2 IS THE OTHER ROOT; DELG IS G-DIFFERENCE
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
      PARAMETER (THIRD = 1.D0/3)
      ZV2 = 0.
      DELG = 0.
      BCN = B*C
      X2 = 1.D0 - BCN
      X1 = A - B*(1.D0+B+C+2*BCN)
      X0 = B*(A-BCN*(1.D0+B))
      Z = X2*THIRD
!
!     START WITH THE VALUE AT THE INFLECTION POINT
!
      F = ((Z-X2)*Z+X1)*Z - X0
      Z = B
      IF ( F.LT.0. .OR. B.GT.Z ) Z = Z + 1
!
!     SET UP THIRD ORDER NEWTON FOR CUBIC EOS
!
  100 CONTINUE
      DF2 = 3.D0*Z - X2
      DF1 = Z*(DF2-X2) + X1
      DF1R = 1.D0/DF1
      F = ((Z-X2)*Z+X1)*Z - X0
      DZ = F*DF1R
      DZ = DZ*(1+DZ*DF2*DF1R)
      Z = Z - DZ
      IF ( ABS(DZ).GT.1.D-7 ) GOTO 100
!
!     CONVERGED; HOW MANY ROOTS ARE DESIRED ?
!
      IF ( MTYP*DF2.GE.0. ) THEN
!
!     FACTOR OUT FIRST ROOT AND CONSIDER REDUCED QUADRATIC
!
         E1 = Z - X2
         E0 = Z*E1 + X1
         D = E1*E1 - 4.D0*E0
!
!     D < 0 MEANS NO MORE ROOTS
!
         IF ( D.GE.0. ) THEN
!
!     GET REMAINING
!
            Z1 = .5D0*(ABS(E1)+SQRT(D))
            IF ( E1.GT.0. ) Z1 = -Z1
            IF ( Z.GT.Z1 ) Z1 = E0/Z1
!
!     REFINE BY A SINGLE NEWTON STEP
!
            DF2 = 3.D0*Z1 - X2
            DF1 = Z1*(DF2-X2) + X1
            F = ((Z1-X2)*Z1+X1)*Z1 - X0
            DF1R = 1.D0/DF1
            DZ = F*DF1R
            DZ = DZ*(1+DZ*DF2*DF1R)
            Z1 = Z1 - DZ
            IF ( Z1.GE.B ) THEN
               IF ( MTYP.EQ.0 ) THEN
!-----CALCULATE EXCESS GIBBS ENERGY DIFFERENCE FOR MULTIPLE SOLUTIONS
                  F = LOG((Z-B)/(Z1-B)) + A/(B*(C2-C1))                 &
     &                *LOG((Z+C2*B)*(Z1+C1*B)/(Z+C1*B)/(Z1+C2*B) )      &
     &                 - (Z-Z1)
                  DELG = ABS(F)
                  ZV2 = Z1
                  IF ( F.LT.0. ) THEN
                     ZV2 = Z
                     Z = Z1
                  ENDIF
               ELSEIF ( MTYP.EQ.1 ) THEN
                  IF ( Z1.LT.Z ) Z = Z1
               ELSE
                  IF ( Z1.GT.Z ) Z = Z1
               ENDIF
            ENDIF
         ENDIF
      ENDIF
      END SUBROUTINE
!
!
      SUBROUTINE TEMSET(T)
!
!     TEMSET CALCULATES THE TEMPERATURE DEPENDENT PART OF THE EOS
!     PURE COMPONENT PARAMETERS, HERE SQRT (A/RT)
!
!     THE RESULTS, THE VECTORS AC0,AC1,AND AC2 REPRESENT
!     VALUE, 1ST DERIVATIVE AND 2ND DERIVATIVE OF SQRT (A/RT)
!
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
      DATA TOLD/0.D0/
      IF ( T.EQ.TOLD ) RETURN
      TOLD = T
      IF (T.EQ. 0) RETURN
      SQT = SQRT(T)
      SQTR = 1.D0/SQT
      T2R = 0.5D0/T
      T2RF = -3*T2R
      DO I = 1 , N
         ALF = TSQR(I)
         Q1 = AC(I)*(1.D0 + Q(I))*SQTR
         AC0(I) = Q1 - AC(I)*Q(I)*ALF
         AC1(I) = -Q1*T2R
         AC2(I) = T2RF*AC1(I)
      ENDDO
      END SUBROUTINE
!
!
!
      SUBROUTINE GETCRIT(I,TCRIT,PCRIT,OMEGA,ITYP,CNAM)
!
!     GETCRIT RETURNS CRITICAL PROPERTIES AND COMPONENT INDICATORS
!
!     I       (I):      COMPONENT INDEX
!     TCRIT   (O):      COMPONENT CRITICAL T
!     PCRIT   (O):      COMPONENT CRITICAL P
!     OMEGA   (O):      COMPONENT ACENTRIC FACTOR
!     ITYP    (O):      COMPONENT TYPE:
!                       SET = 0 FOR INTERMEDIATE HYDROCARBONS,
!                       1 FOR OTHERS
!                       (ONLY USED IN MULTIPHASE FLASH)
!
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
      CHARACTER*4 CNAM
      TCRIT = TCR(I)
      PCRIT = PCR(I)
      OMEGA = OMG(I)
      ITYP   = NOTRY(I)
      CNAM = NWNAME(I)
      END SUBROUTINE
!
!
!
      SUBROUTINE INI_CUB(NC,ICEQ,LIST)
!
!     A SMALL 'DATABASE ROUTINE' FOR PROPERTIES INPUT
!
!     NC:     (I):      NO. OF COMPONENT
!     ICEQ:   (I):      EOS; 0 = SRK, 1 = PENG-ROBINSON
!     LIST:   (I):      VECTOR OF COMPONENT INDICES
!
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
      PARAMETER (MAXC=15)
      DIMENSION LIST(*)
      CHARACTER (LEN=4) NAME(MAXC)
      REAL T(MAXC),P(MAXC),OMEGA(MAXC),CKV(MAXC,MAXC)
      DIMENSION NHYCIF(MAXC)
!-------EOS CONSTANTS

!-----IDENTIFY HYDROCARBONS
      DATA NHYCIF/11*0,2,3*1/
!-----IDEAL GAS HEAT CAPACITY POLYNOMIAL COEFFICIENTS (REID ET AL.)
      DATA NAME/                                                        &
     &      ' C1 ',' C2 ',' C3 ',' IC4',' C4 ',                         &
     &      ' IC5',' C5 ',' C6 ',' C7 ',' C8 ',                         &
     &      ' C9 ',' H20',' N2 ',' C02',' H2S'/
!------CRITICAL TEMPERATURES (K) AND PRESSURES (MPa) (REID ET AL.)
      DATA T/                                                           &
     &   190.6, 305.4, 369.8, 408.1, 425.2,                             &
     &   460.4, 469.6, 507.4, 540.2, 568.8,                             &
     &   594.6, 647.3, 126.2, 304.2, 373.2/
      DATA P/                                                           &
     &     4.599,  4.8827, 4.2445, 3.6468, 3.7988,                      &
     &     3.3834, 3.3733, 2.9681, 2.7351, 2.4819,                      &
     &     2.3096,22.0834, 3.3936, 7.3746, 8.9374/

!------ACENTRIC FACTORS (REID ET AL.)
      DATA OMEGA/                                                       &
     &     0.008, 0.098, 0.152, 0.176, 0.193,                           &
     &     0.227, 0.251, 0.296, 0.351, 0.394,                           &
     &     0.440, 0.344, 0.040, 0.225, 0.100/
!------NONZERO KIJ FOR SRK-EQUATION (REID ET AL., HEIDEMANN ET AL.)
      DATA (CKV( 1,J),J= 2,15)/10*0.,.45,.02,.12,.08/
      DATA (CKV( 2,J),J= 3,15)/ 9*0.,.45,.06,.15,.07/
      DATA (CKV( 3,J),J= 4,15)/ 8*0.,.53,.08,.15,.07/
      DATA (CKV( 4,J),J= 5,15)/ 7*0.,.52,.08,.15,.06/
      DATA (CKV( 5,J),J= 6,15)/ 6*0.,.52,.08,.15,.06/
      DATA (CKV( 6,J),J= 7,15)/ 5*0.,.50,.08,.15,.06/
      DATA (CKV( 7,J),J= 8,15)/ 4*0.,.50,.08,.15,.06/
      DATA (CKV( 8,J),J= 9,15)/ 3*0.,.50,.08,.15,.05/
      DATA (CKV( 9,J),J=10,15)/ 2*0.,.50,.08,.15,.04/
      DATA (CKV(10,J),J=11,15)/   0.,.50,.08,.15,.04/
      DATA (CKV(11,J),J=12,15)/      .50,.08,.15,.03/
      DATA (CKV(12,J),J=13,15)/                 3*0./
      DATA (CKV(13,J),J=14,15)/                 2*0./
      DATA  CKV(14,15)        /                  .12/
      DO K=1,15
         CKV(K,K) = 0.
      ENDDO
      DO K=1,14
         DO I=K+1,15
              CKV(I,K) = CKV(K,I)
         ENDDO
      ENDDO
!
!     GET EOS CONSTANTS
!
      IF (ICEQ .EQ. 0) THEN
         C = 0.D0
         ELSE
         C = 1.D0
      ENDIF
      CALL EOS_CONSTANTS(C,C1,C2,CONA,CONB)
      N = NC
!
!     CLEAR OLD SPACE
!
      IF (ALLOCATED (NOTRY )) DEALLOCATE(NOTRY)
      IF (ALLOCATED (AC    )) DEALLOCATE(AC   )
      IF (ALLOCATED (Q     )) DEALLOCATE(Q    )
      IF (ALLOCATED (TSQR  )) DEALLOCATE(TSQR )
      IF (ALLOCATED (TCR   )) DEALLOCATE(TCR  )
      IF (ALLOCATED (PCR   )) DEALLOCATE(PCR  )
      IF (ALLOCATED (OMG   )) DEALLOCATE(OMG  )
      IF (ALLOCATED (BC    )) DEALLOCATE(BC   )
      IF (ALLOCATED (AC0   )) DEALLOCATE(AC0  )
      IF (ALLOCATED (AC1   )) DEALLOCATE(AC1  )
      IF (ALLOCATED (AC2   )) DEALLOCATE(AC2  )
      IF (ALLOCATED (ONEMK )) DEALLOCATE(ONEMK)
      IF (ALLOCATED (NWNAME)) DEALLOCATE(NWNAME)
!
!     ALLOCATE SPACE
!
      ALLOCATE (NOTRY(N),TCR(N),PCR(N),OMG(N),BC(N))
      ALLOCATE (TSQR(N),Q(N),AC(N))
      ALLOCATE (AC0(N),AC1(N),AC2(N),ONEMK(N,N),NWNAME(N))
!
!    CLEAR TEMSET-PARAMETERS
!
      CALL TEMSET(0.D0)
!
!     RUN THROUGH LIST ELEMENTS TO SELECT COMPONENTS
!
      DO I=1,N
         L=LIST(I)
         NOTRY(I)=NHYCIF(L)
         NWNAME(I)=NAME(L)
!
!     CRITICAL TEMPERATURE AND PRESSURE
!
         TCR(I)=T(L)
         TSQR(I)=1.D0/SQRT(TCR(I))
         PCR(I)=P(L)
!
!     EOS CHARACTERISTIC CONSTANTS
!
         BC(I)=CONB*TCR(I)/PCR(I)
         AC(I)=CONA*TCR(I)/SQRT(PCR(I))
         OMG(I)=OMEGA(L)
         IF (ICEQ .EQ.0) THEN
!-----ICEQ=0: SRK
             Q(I)=.48+OMG(I)*(1.574-.176*OMG(I))
             ELSE
!-----ICEQ=1: PENG-ROBINSON
             Q(I)=.37464+OMG(I)*(1.54226-.26992*OMG(I))
         ENDIF
      ENDDO
      CRMAX=0.
      CRMIN=1.D5
      LMAX=0
      LMIN=0
!------CHOOSE LIGHT AND HEAVY HYDROCARB. FROM CRIT. TEMP
!
!     NOTRY = 0  IS THE MARK FOR HYDROCARBONS
!
      DO I=1,N
         IF (NOTRY(I).LT.1) THEN
            TTRY=TCR(I)
            IF (TTRY.GT. CRMAX) THEN
                CRMAX=TTRY
                LMAX=I
            ENDIF
            IF (TTRY.LT. CRMIN) THEN
                CRMIN=TTRY
                LMIN=I
            ENDIF
         ENDIF
      ENDDO
      IF (LMIN.NE.0) NOTRY(LMIN)=1
      IF (LMAX.NE.0) NOTRY(LMAX)=1
      DO I=1,N
         DO J=I,N
            ONEMK(I,J)=1.D0-CKV(LIST(I),LIST(J))
            ONEMK(J,I)=ONEMK(I,J)
         ENDDO
      ENDDO
      END SUBROUTINE
      SUBROUTINE EOS_CONSTANTS (CPAR,DEL1,DEL2,CONA,CONB)
!
!     THIS SUBROUTINE CALCULATES THE EOS-CONSTANTS FOR A CUBIC EOS OF
!     THE FORM
!
!     NOTE THAT CONB = OMEGA_B, BUT CONA = SQRT (OMEGA_A)
!
!     P = RT/(v-b) - a/(v^2 + (c+1)vb - cb^2)  =
!       = RT/(v-b) - a/((v+del1 b)*(v+del2 b))
!
!     where:
!
!     c = 0:  Redlich-Kwong
!     c = 1:  Peng-Robinson
!
!     The coding is based on Michelsen and Mollerup, Ch. 3,
!     section 3.2
!
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
      ROOT = SQRT ( (CPAR+1.D0)**2 + 4.D0*CPAR)
      DEL1 = 0.50D0*(CPAR + 1.D0 + ROOT)
      DEL2 = -CPAR/DEL1
      D1 = 1.D0 + CPAR
      D2 = -CPAR
      FAC_1  =  (1.D0+DEL1)**(1.D0/3)
      FAC_2  =  (1.D0+DEL2)**(1.D0/3)
      APLUSB = FAC_1*FAC_2*(FAC_1+FAC_2)
      Y = 1.D0 + APLUSB
      OMEG_RAT =(3.D0*Y*(Y+D1) + D1*D1 - D2)/(3.D0*Y+CPAR)
      Z_CRIT = Y/(3.D0*Y+CPAR)
      OMEG_B = Z_CRIT/Y
      OMEG_A = OMEG_RAT*OMEG_B
      CONB = OMEG_B
      CONA = SQRT (OMEG_A)
      END SUBROUTINE
      END
